﻿using System;
using System.Net;
using System.Text;
using System.Text.RegularExpressions;
using System.Linq;
using System.Linq.Expressions;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Reflection;
using System.Threading;
using System.Diagnostics;
using System.ComponentModel;

using Zorbo.Resources;
using Zorbo.Sockets;

using Zorbo.Packets;
using Zorbo.Packets.Ares;

using Zorbo.Interface;
using Zorbo.Users;

namespace Zorbo
{
    #region " Counter "

    public class Counter
    {
        public int Count { get; set; }
        public DateTime Last { get; set; }

        public Counter() { }

        public Counter(int count, DateTime last) {
            Last = last;
            Count = count;
        }
    }

    #endregion

    public class AresClient : IClient, IDisposable
    {
        #region " Variables "

        ISocket socket;

        AresServer server;
        AdminLevel admin;

        ushort id;
        Guid guid;

        uint cookie;

        bool local;
        bool browse;
        bool cloaked;
        bool muzzled;
        bool captcha;
        bool fastping;
        bool compression;

        bool logged;

        byte age;
        byte sex;
        byte country;

        IAvatar avatar;
        IAvatar orgavatar;

        string name;
        string orgname;
        string version;
        string region;
        string message;

        ushort vroom;
        ushort dcport;
        ushort nodeport;

        IPAddress nodeip;
        IPAddress localip;

        IPHostEntry dnsentry;

        DateTime spawntime;
        DateTime captchatime;
        DateTime lastupdate;

        List<String> ignored;

        Dictionary<Byte, Counter> counters;

        ObservableCollection<ISharedFile> files;
        ReadOnlyList<ISharedFile> pfiles;

        Dictionary<String, Object> extended_props;

        int captchaTries = 5;
        int captchaAnswer;

        #endregion

        #region " Properties "

        public ushort Id { 
            get { return id; }
            private set { id = value; }
        }


        public Guid Guid { 
            get { return guid; }
            private set { guid = value; }
        }

        public uint Cookie { 
            get {
                if (cookie == 0)
                    cookie = (uint)GetHashCode();

                return cookie; 
            }
        }

        public ISocket Socket {
            get { return socket; }
        }

        public bool IsCaptcha {
            get { return captcha; }
            set {
                if (captcha != value) {
                    captcha = value;
                    RaisePropertyChanged(() => IsCaptcha);

                    var record = server.History.Find((s) => s.Equals(this));

                    if (value) {
                        record.Trusted = false;

                        if (LoggedIn)
                            server.SendPacket((s) => 
                                s != this &&
                                s.Vroom == Vroom, 
                                new Parted(Name));

                        ShowCaptcha();
                    }
                    else {
                        captchaTries = 5;

                        record.Trusted = true;
                        server.PluginManager.OnCaptcha(this, CaptchaEvent.Exit);

                        PerformLogin();

                        server.PluginManager.OnJoin(this);
                    }
                }
            }
        }

        public bool LoggedIn {
            get { return logged; }
            private set { logged = value; }
        }

        public bool Connected {
            get { return socket != null && socket.Connected; }
        }

        public bool LocalHost {
            get { return local; }
            internal set {
                if (value) {
                    local = value;
                    RaisePropertyChanged(() => LocalHost);
                }
            }
        }

        public bool Browsable {
            get { return browse; }
            private set {
                if (browse != value) {
                    browse = value;
                    RaisePropertyChanged(() => Browsable);
                }
            }
        }

        public bool Cloaked {
            get { return cloaked; }
            set {
                if (cloaked != value) {
                    cloaked = value;
                    RaisePropertyChanged(() => Cloaked);

                    if (cloaked)
                        server.SendPacket((s) =>
                            s != this &&
                            s.Admin < Admin &&
                            s.Vroom == Vroom,
                            new Parted(Name));
                    else
                        server.SendPacket((s) =>
                            s != this &&
                            s.Admin < Admin &&
                            s.Vroom == Vroom,
                            new Join(this));
                }
            }
        }

        public bool Muzzled {
            get { return muzzled; }
            set {
                if (muzzled != value) {
                    muzzled = value;
                    RaisePropertyChanged(() => Muzzled);

                    if (muzzled)
                        socket.SendAsync(new Announce(Strings.Muzzled));
                    else
                        socket.SendAsync(new Announce(Strings.Unmuzzled));
                }
            }
        }

        public bool FastPing {
            get { return fastping; }
            private set { fastping = value; }
        }

        public bool Compression {
            get { return compression; }
            private set {
                if (compression != value) {
                    compression = value;
                    RaisePropertyChanged(() => Compression);
                }
            }
        }



        public byte Age {
            get { return age; }
            private set {
                if (age != value) {
                    age = value;
                    RaisePropertyChanged(() => Age);
                }
            }
        }

        public byte Sex { 
            get { return sex; }
            private set {
                if (sex != value) {
                    sex = value;
                    RaisePropertyChanged(() => Sex);
                }
            }
        }

        public byte Country { 
            get { return country; }
            private set {
                if (country != value) {
                    country = value;
                    RaisePropertyChanged(() => Country);
                }
            }
        }

        public AdminLevel Admin {
            get { return admin; }
            set {
                if (admin != value) {

                    if (LocalHost && value < admin)
                        return;

                    admin = value;
                    RaisePropertyChanged(() => Admin);

                    socket.SendAsync(new OpChange(Admin > 0));

                    server.SendPacket((s) =>
                        s.Vroom == Vroom,
                        new ServerUpdate(this));
                }
            }
        }


        public IAvatar Avatar {
            get {
                if (avatar == null)
                    return server.Config.Avatar;

                return avatar; 
            }
            set {
                if (!value.Equals(AresAvatar.Null) &&
                    avatar == null || !avatar.Equals(value)) {

                    avatar = value;
                    RaisePropertyChanged(() => Avatar);

                    server.SendPacket((s) =>
                        s.Vroom == Vroom,
                        new ServerAvatar(this));
                }
            }
        }

        public IAvatar OrgAvatar { 
            get { return orgavatar; }
            private set { orgavatar = value; }
        }


        public string Name {
            get { return name; }
            set {
                if (name != value)
                    ChangeName(name, value);
            }
        }

        public string OrgName {
            get { return orgname; }
            private set { orgname = value; }
        }

        public string Version { 
            get { return version; }
            private set { version = value; }
        }

        public string Region { 
            get { return region; }
            internal set {
                if (region != value) {
                    region = value;
                    RaisePropertyChanged(() => Region);
                }
            }
        }

        public string Message { 
            get { return message; }
            set {
                if (message != value) {
                    message = value;
                    RaisePropertyChanged(() => Message);

                    server.SendPacket((s) =>
                        s.Vroom == Vroom,
                        new ServerPersonal(Name, message));
                }
            }
        }


        public ushort Vroom {
            get { return vroom; }
            set {
                if (vroom != value)
                    ChangeVroom(vroom, value);
            }
        }

        public ushort FileCount {
            get { return (ushort)files.Count; }
        }

        public ushort ListenPort { 
            get { return dcport; }
            private set { dcport = value; }
        }

        public ushort NodePort {
            get { return nodeport; }
            private set { nodeport = value; }
        }


        public IPAddress NodeIp {
            get { return nodeip; }
            private set {
                if (nodeip != value) {
                    nodeip = value;
                    RaisePropertyChanged(() => NodeIp);
                }
            }
        }

        public IPAddress LocalIp {
            get { return localip; }
            private set {
                if (localip != value) {
                    localip = value;
                    RaisePropertyChanged(() => LocalIp);
                }
            }
        }

        public IPAddress ExternalIp {
            get {
                if (socket != null)
                    try {
                        return socket.RemoteEndPoint.Address;
                    }
                    catch { }

                return null;
            }
        }

        public IPHostEntry DnsEntry {
            get { return dnsentry; }
            internal set {
                if (dnsentry == null || !dnsentry.Equals(value)) {
                    dnsentry = value;
                    RaisePropertyChanged(() => DnsEntry);
                }
            }
        }


        public IList<String> Ignored {
            get { return ignored; }
        }

        public IReadOnlyList<ISharedFile> Files {
            get { return pfiles; }
        }


        IDictionary<String, Object> IClient.Extended {
            get { return extended_props; }
        }


        public DateTime SpawnTime {
            get { return spawntime; }
            internal set { spawntime = value; }
        }

        public DateTime CaptchaTime {
            get { return captchatime; }
            internal set { captchatime = value; }
        }

        public DateTime LastUpdate {
            get { return lastupdate; }
            internal set { lastupdate = value; }
        }

        

        internal Dictionary<Byte, Counter> Counters {
            get { return counters; }
            private set { counters = value; }
        }

        #endregion


        internal AresClient(AresServer server, ISocket socket, ushort id) {
            this.id = id;

            this.server = server;
            this.socket = socket;

            this.spawntime = TimeBank.CurrentTime;
            this.lastupdate = this.spawntime;
            
            this.ignored = new List<String>();

            this.files = new ObservableCollection<ISharedFile>();
            this.pfiles = new ReadOnlyList<ISharedFile>(files);

            this.counters = new Dictionary<Byte, Counter>();
            this.extended_props = new Dictionary<String, Object>();

            this.server.Config.PropertyChanged += Config_PropertyChanged;
        }


        void Config_PropertyChanged(object sender, PropertyChangedEventArgs e) {

            if (e.PropertyName == "Avatar") {
                if (avatar == null)
                    server.SendPacket((s) => s.Vroom == Vroom, new ServerAvatar(Name, server.Config.Avatar));

                socket.SendAsync(new ServerAvatar("Zorbo", server.Config.Avatar));
            }
        }
        


        private void ShowCaptcha() {

            if (!LoggedIn) {
                LoggedIn = true;
                PerformLogin();
            }
            else PerformQuickLogin();

            captchatime = TimeBank.CurrentTime;
            captchaAnswer = Captcha.Create(this);
        }

        private void FinishCaptcha(string message) {
            int sum = 0;

            if (Int32.TryParse(message, out sum)) {
                if (sum == captchaAnswer) {
                    IsCaptcha = false;
                    return;
                }
            }

            if (--captchaTries <= 0) {
                socket.SendAsync(new Announce(Strings.BannedCaptchaAnswers));
                server.PluginManager.OnCaptcha(this, CaptchaEvent.Banned);

                Ban();
            }
            else {
                socket.SendAsync(new Announce(String.Format(Strings.InvalidCaptcha, captchaTries)));
                server.PluginManager.OnCaptcha(this, CaptchaEvent.Failed);
            }
        }


        internal void PerformLogin() {
            socket.SendAsync(new LoginAck() {
                Username = Name,
                ServerName = server.Config.Name,
            });

            socket.SendAsync(new Features() {
                Version = Strings.VersionLogin,
                SupportFlag = 3,
                SharedTypes = 63,
                Language = server.Config.Language,
                Cookie = this.Cookie,
            });

            socket.SendAsync(new TopicFirst(server.Config.Topic));
            server.SendUserlist(this);

            if (!IsCaptcha) {
             
                server.SendAvatars(this);
                socket.SendAsync(new OpChange(Admin > AdminLevel.User));

                Thread.Sleep(500);

                server.PluginManager.OnSendLogin(this);
            }
            else server.PluginManager.OnCaptcha(this, CaptchaEvent.Enter);
        }

        internal void PerformQuickLogin() {

            socket.SendAsync(new LoginAck() {
                Username = Name,
                ServerName = server.Config.Name,
            });

            socket.SendAsync(new TopicFirst(server.Config.Topic));
            server.SendUserlist(this);

            if (!IsCaptcha) {
                if (!String.IsNullOrEmpty(Message))
                    server.SendPacket((s) =>
                        s.Vroom == Vroom,
                        new ServerPersonal(Name, Message));

                server.SendAvatars(this);
                socket.SendAsync(new OpChange(Admin > 0));

                Thread.Sleep(500);

                server.PluginManager.OnSendLogin(this);
            }
            else server.PluginManager.OnCaptcha(this, CaptchaEvent.Enter);
        }


        internal void ChangeName(string oldname, string newname) {
            name = newname.FormatUsername();
            RaisePropertyChanged(() => Name);

            server.SendPacket((s) =>
                s != this && 
                s.Vroom == Vroom,
                new Parted() { Username = oldname });

            PerformQuickLogin();
        }

        internal void ChangeVroom(ushort oldvroom, ushort newvroom) {

            if (server.PluginManager.OnVroomJoinCheck(this, newvroom)) {

                server.PluginManager.OnVroomPart(this);

                vroom = newvroom;
                RaisePropertyChanged(() => Vroom);

                server.SendPacket((s) =>
                    s != this &&
                    s.Vroom == oldvroom,
                    new Parted(Name));

                PerformQuickLogin();

                server.SendAnnounce(this, String.Format(Strings.VroomNotice, newvroom));
                server.PluginManager.OnVroomJoin(this);
            }
        }


        internal void HandleJoin(PacketEventArgs e) {
            Login login = (Login)e.Packet;

            foreach (var user in server.Users)
                if (user.Guid.Equals(login.Guid))
                    user.Disconnect();

            Guid = login.Guid;
            ListenPort = login.ListenPort;
            NodeIp = login.NodeIp;
            NodePort = login.NodePort;
            name = login.Username.FormatUsername();
            orgname = name;
            Version = login.Version;
            LocalIp = login.LocalIp;
            Browsable = (login.SupportFlag & 2) == 2;
            Compression = (login.SupportFlag & 4) == 4;
            Age = login.Age;
            Sex = login.Sex;
            Country = login.Country;
            Region = login.Region;

            var record = server.History.Add(this);
            var autologin = server.History.Admin.Passwords.Find((s) => s.Id.Equals(record));

            admin = (autologin != null) ? autologin.Level : admin;
            admin = (LocalHost) ? AdminLevel.Host : admin;

            if (admin != AdminLevel.User) RaisePropertyChanged("Admin");

            if (!LocalHost) { //wait for Dns, shouldn't take long work or fail..

                DnsHelper.Resolve(ExternalIp, (ip, entry) => {
                    if (entry != null) DnsEntry = entry;

                    if (AllowedJoin(record)) {

                        if (server.Config.BotProtection) {
                            if (!record.Trusted) {
                                IsCaptcha = true;
                                return;
                            }
                        }

                        FinishJoin();
                    }
                });
            }
            else if (AllowedJoin(record))
                FinishJoin();
        }

        private void FinishJoin() {

            LoggedIn = true;
            PerformLogin();

            server.PluginManager.OnJoin(this);
        }

        internal bool HandlePacket(PacketEventArgs e) {
            if (!LoggedIn)
                Console.WriteLine("Client tried sending packet before they we're fully logged in");
            
            if (IsCaptcha) {
                switch ((AresId)e.Packet.Id) {
                    case AresId.MSG_CHAT_CLIENT_FASTPING:
                        fastping = true;
                        break;
                    case AresId.MSG_CHAT_CLIENT_AUTOLOGIN: {
                            AutoLogin login = (AutoLogin)e.Packet;
                            var pass = server.History.Admin.Passwords.CheckSha1(this, login.Sha1Password);

                            if (pass == null)
                                socket.SendAsync(new Announce(Strings.InvalidLogin));
                            else
                                socket.SendAsync(new Announce(String.Format(Strings.LoggedIn, pass.Level)));
                        }
                        break;
                    case AresId.MSG_CHAT_CLIENT_PUBLIC:
                        ClientPublic pub = (ClientPublic)e.Packet;
                        FinishCaptcha(pub.Message);
                        break;
                    case AresId.MSG_CHAT_CLIENT_EMOTE:
                        ClientEmote emote = (ClientEmote)e.Packet;
                        FinishCaptcha(emote.Message);
                        break;
                    case AresId.MSG_CHAT_CLIENT_ADDSHARE:
                        SharedFile addfile = (SharedFile)e.Packet;

                        if (server.PluginManager.OnFileReceived(this, addfile))
                            lock (files) files.Add(addfile);

                        break;
                    case AresId.MSG_CHAT_CLIENT_UPDATE_STATUS:
                        ClientUpdate update = (ClientUpdate)e.Packet;

                        lastupdate = TimeBank.CurrentTime;

                        NodeIp = update.NodeIp;
                        NodePort = update.NodePort;
                        Age = (update.Age != 0) ? update.Age : Age;
                        Sex = (update.Sex != 0) ? update.Sex : Sex;
                        Country = (update.Country != 0) ? update.Country : Country;
                        Region = !String.IsNullOrEmpty(update.Region) ? update.Region : Region;
                        break;
                }

                return true; //don't handle any other packets yet
            }
            else if (LoggedIn) {

                switch ((AresId)e.Packet.Id) {
                    case AresId.MSG_CHAT_CLIENT_FASTPING:
                        fastping = true;
                        return true;
                    case AresId.MSG_CHAT_CLIENT_DUMMY:
                        return true;
                    case AresId.MSG_CHAT_CLIENT_PUBLIC:
                        ClientPublic pub = (ClientPublic)e.Packet;

                        if (!String.IsNullOrEmpty(pub.Message))
                            if (pub.Message.StartsWith("#") && Commands.HandlePreCommand(server, this, pub.Message.Substring(1)))
                                return true;

                        if (Muzzled) {
                            socket.SendAsync(new Announce(Strings.AreMuzzled));
                            return true;
                        }

                        return false;
                    case AresId.MSG_CHAT_CLIENT_EMOTE:
                        ClientEmote emote = (ClientEmote)e.Packet;

                        if (!String.IsNullOrEmpty(emote.Message))
                            if (emote.Message.StartsWith("#") && Commands.HandlePreCommand(server, this, emote.Message.Substring(1)))
                                return true;

                        if (Muzzled) {
                            socket.SendAsync(new Announce(Strings.AreMuzzled));
                            return true;
                        }

                        return false;
                    case AresId.MSG_CHAT_CLIENT_COMMAND:
                        Command cmd = (Command)e.Packet;
                        Commands.HandlePreCommand(server, this, cmd.Message);
                        break;
                    case AresId.MSG_CHAT_CLIENT_PVT:
                        Private pvt = (Private)e.Packet;

                        if (Muzzled && !server.Config.MuzzledPMs) {

                            pvt.Message = "[" + Strings.AreMuzzled + "]";
                            socket.SendAsync(pvt);

                            return true;
                        }

                        return false;
                    case AresId.MSG_CHAT_CLIENT_AUTHREGISTER: {
                            AuthRegister reg = (AuthRegister)e.Packet;
                            Commands.HandleRegister(server, this, reg.Password);
                        }
                        return true;
                    case AresId.MSG_CHAT_CLIENT_AUTHLOGIN: {
                            AuthLogin login = (AuthLogin)e.Packet;
                            Commands.HandleLogin(server, this, login.Password);
                        }
                        return true;
                    case AresId.MSG_CHAT_CLIENT_AUTOLOGIN: {
                            AutoLogin login = (AutoLogin)e.Packet;
                            var pass = server.History.Admin.Passwords.CheckSha1(this, login.Sha1Password);

                            if (pass == null)
                                socket.SendAsync(new Announce(Strings.InvalidLogin));
                            else
                                socket.SendAsync(new Announce(String.Format(Strings.LoggedIn, pass.Level)));
                        }
                        return true;
                    case AresId.MSG_CHAT_CLIENT_ADDSHARE:
                        SharedFile addfile = (SharedFile)e.Packet;

                        if (server.PluginManager.OnFileReceived(this, addfile))
                            lock (files) files.Add(addfile);

                        return true;
                    case AresId.MSG_CHAT_CLIENT_IGNORELIST:
                        Ignored ignore = (Ignored)e.Packet;

                        lock (ignored) {
                            if (!ignored.Contains(ignore.Username)) {

                                ignored.Add(ignore.Username);
                                socket.SendAsync(new Announce(String.Format(Strings.Ignored, ignore.Username)));
                            }
                            else {
                                ignored.Remove(ignore.Username);
                                socket.SendAsync(new Announce(String.Format(Strings.Ignored, ignore.Username)));
                            }
                        }

                        return true;
                    case AresId.MSG_CHAT_CLIENT_UPDATE_STATUS:
                        ClientUpdate update = (ClientUpdate)e.Packet;

                        lastupdate = TimeBank.CurrentTime;

                        NodeIp = update.NodeIp;
                        NodePort = update.NodePort;
                        Age = (update.Age != 0) ? update.Age : Age;
                        Sex = (update.Sex != 0) ? update.Sex : Sex;
                        Country = (update.Country != 0) ? update.Country : Country;
                        Region = !String.IsNullOrEmpty(update.Region) ? update.Region : Region;

                        server.SendPacket((s) => s.Vroom == Vroom, new ServerUpdate(this));
                        return true;
                    case AresId.MSG_CHAT_CLIENT_DIRCHATPUSH:
                        ClientDirectPush push = (ClientDirectPush)e.Packet;

                        if (Encoding.UTF8.GetByteCount(push.Username) < 2) {
                            socket.SendAsync(new DirectPushError(4));
                            return true;
                        }

                        if (push.TextSync.Length < 16) {
                            socket.SendAsync(new DirectPushError(3));
                            return true;
                        }

                        IClient target = server.FindUser(s => s.Name == push.Username);

                        if (target == null) {
                            socket.SendAsync(new DirectPushError(1));
                            return true;
                        }

                        if (target.Ignored.Contains(Name)) {
                            socket.SendAsync(new DirectPushError(2));
                            return true;
                        }

                        socket.SendAsync(new DirectPushError(0));
                        target.Socket.SendAsync(new ServerDirectPush(this, push));

                        return true;
                    case AresId.MSG_CHAT_CLIENT_BROWSE:
                        Browse browse = (Browse)e.Packet;

                        browse.Type = (byte)((browse.Type == 0) ? (byte)255 : browse.Type);
                        browse.Type = (byte)((browse.Type == 8) ? (byte)0 : browse.Type);

                        IClient browse_target = server.FindUser(s => s.Vroom == Vroom && s.Name == browse.Username);

                        if (browse_target == null)
                            return true;

                        else if (browse_target.Files.Count == 0)
                            socket.SendAsync(new BrowseError(browse.BrowseId));

                        else {
                            socket.SendAsync(new BrowseStart(browse.BrowseId, (ushort)browse_target.Files.Count));

                            foreach (var file in browse_target.Files) {
                                if (browse.Type == 255 || browse.Type == file.Type)
                                    socket.SendAsync(new BrowseItem(browse.BrowseId, file));
                            }

                            socket.SendAsync(new BrowseEnd(browse.BrowseId));
                        }

                        return true;
                    case AresId.MSG_CHAT_CLIENT_SEARCH:
                        Search search = (Search)e.Packet;

                        search.Type = (byte)((search.Type == 0) ? (byte)255 : search.Type);
                        search.Type = (byte)((search.Type == 8) ? (byte)0 : search.Type);

                        foreach (var user in server.Users) {
                            if (user != this && user.Vroom == Vroom) {

                                foreach (var file in user.Files) {
                                    if (file.SearchWords.ContainsAny(search.SearchWords.Split(new char[] { '\0' }, StringSplitOptions.RemoveEmptyEntries)))
                                        socket.SendAsync(new SearchHit(search.SearchId, user, file));
                                }
                            }
                        }

                        socket.SendAsync(new SearchEnd(search.SearchId));
                        return true;
                }

                return false;
            }
            else return true;
        }

        internal void HandleOverridePacket(PacketEventArgs e) {

            switch ((AresId)e.Packet.Id) {

                case AresId.MSG_CHAT_CLIENT_PUBLIC:
                    ClientPublic pub = (ClientPublic)e.Packet;

                    if (!String.IsNullOrEmpty(pub.Message)) {

                        if (pub.Message.StartsWith("#"))
                            Commands.HandleCommand(server, this, pub.Message.Substring(1));

                        server.SendPacket((s) =>
                            s.Vroom == Vroom &&
                           !s.Ignored.Contains(Name), 
                            new ServerPublic(Name, pub.Message));
                    }
                    break;
                case AresId.MSG_CHAT_CLIENT_EMOTE:
                    ClientEmote emote = (ClientEmote)e.Packet;

                    if (!String.IsNullOrEmpty(emote.Message)) {

                        if (emote.Message.StartsWith("#"))
                            Commands.HandleCommand(server, this, emote.Message.Substring(1));

                        server.SendPacket((s) => 
                            s.Vroom == Vroom &&
                           !s.Ignored.Contains(Name), 
                            new ServerEmote(Name, emote.Message));
                    }
                    break;
                case AresId.MSG_CHAT_CLIENT_PVT: {
                        Private priv = (Private)e.Packet;

                        if (String.IsNullOrEmpty(priv.Message)) return;

                        IClient target = server.FindUser((s) => s.Name == priv.Username);

                        if (target != null) {

                            if (target.Ignored.Contains(Name))
                                socket.SendAsync(new IgnoringYou(priv.Username));
                            else {
                                priv.Username = Name;
                                target.Socket.SendAsync(priv);
                            }
                        }
                        else socket.SendAsync(new Offline() { Username = priv.Username });
                    }
                    break;
                case AresId.MSG_CHAT_CLIENT_COMMAND:
                    Command cmd = (Command)e.Packet;
                    Commands.HandleCommand(server, this, cmd.Message);
                    break;
                case AresId.MSG_CHAT_CLIENT_PERSONAL_MESSAGE:
                    ClientPersonal personal = (ClientPersonal)e.Packet;
                    Message = personal.Message;

                    break;
                case AresId.MSG_CHAT_CLIENT_AVATAR:
                    ClientAvatar avatar = (ClientAvatar)e.Packet;

                    if (Avatar.Equals(avatar))
                        break;

                    if (avatar.AvatarBytes.Length == 0)
                        Avatar = AresAvatar.Null;
                    else {
                        Avatar = new AresAvatar(avatar.AvatarBytes);

                        if (OrgAvatar == null)
                            OrgAvatar = new AresAvatar(avatar.AvatarBytes);
                    }
                    break;
                case AresId.MSG_CHAT_CLIENT_CUSTOM_DATA: {
                        ClientCustom custom = (ClientCustom)e.Packet;

                        string username = custom.Username;
                        custom.Username = Name;

                        IClient target = server.FindUser((s) => s.Name == username);

                        if (target != null && !target.Ignored.Contains(Name)) 
                            target.Socket.SendAsync(custom);
                    }
                    break;
                case AresId.MSG_CHAT_CLIENT_CUSTOM_DATA_ALL:
                    ClientCustomAll customAll = (ClientCustomAll)e.Packet;
                    
                    server.SendPacket((s) =>
                        s != this &&
                        s.Vroom == Vroom,
                        new ClientCustom(Name, customAll));

                    break;
            }
        }


        private bool AllowedJoin(IRecord record) {
            try {
                if (this.Guid.Equals(Guid.Empty)) {

                    socket.SendAsync(new Announce(Errors.InvalidLogin));
                    server.PluginManager.OnJoinRejected(this, RejectReason.InvalidLogin);

                    Disconnect();
                    return false;
                }
                    
                //let onjoincheck run for host but ignore its return
                if (!server.PluginManager.OnJoinCheck(this) && !LocalHost) {

                    socket.SendAsync(new Announce(Errors.Rejected));
                    server.PluginManager.OnJoinRejected(this, RejectReason.Plugins);

                    Disconnect();
                    return false;
                }

                if (!LocalHost) {

                    if (server.History.Bans.Contains((s) => s.Equals(record))) {

                        socket.SendAsync(new Error(Errors.Banned));
                        server.PluginManager.OnJoinRejected(this, RejectReason.Banned);

                        Disconnect();
                        return false;
                    }

                    if (server.History.RangeBans.Contains((s) => s.IsMatch(ExternalIp.ToString()))) {

                        socket.SendAsync(new Error(Errors.RangeBanned));
                        server.PluginManager.OnJoinRejected(this, RejectReason.RangeBanned);

                        Disconnect();
                        return false;
                    }

                    if (DnsEntry != null && server.History.DnsBans.Contains((s) => s.IsMatch(DnsEntry.HostName))) {

                        socket.SendAsync(new Error(Errors.DnsBanned));
                        server.PluginManager.OnJoinRejected(this, RejectReason.DnsBanned);

                        Disconnect();
                        return true;
                    }
                }

                int count = 1;
                foreach (var user in server.Users) {
                    if (user != this) {

                        if (user.Name == Name) {

                            socket.SendAsync(new Announce(Errors.NameTaken));
                            server.PluginManager.OnJoinRejected(this, RejectReason.NameTaken);

                            Disconnect();
                            return false;
                        }
                        else if (user.ExternalIp.Equals(socket.RemoteEndPoint.Address))
                            count++;
                    }

                    if (count > server.Config.MaxClones) {

                        socket.SendAsync(new Error(String.Format(Errors.Clones, socket.RemoteEndPoint.Address)));
                        server.PluginManager.OnJoinRejected(this, RejectReason.TooManyBots);

                        Disconnect();
                        return false;
                    }
                }

                return true;
            }
            catch (Exception) { return false; }
        }


        public void Ban() {
            Ban(null);
        }

        public void Ban(Object state) {
            server.History.Bans.Add(this);
            Disconnect(state);
        }


        public void Disconnect() {
            Disconnect(null);
        }

        public void Disconnect(Object state) {

            if (LoggedIn) {
                logged = false;

                server.History.Add(this);

                server.SendPacket((s) => 
                    s.Vroom == vroom, 
                    new Parted(name));
            }

            if (socket != null)
                socket.Disconnect(state);
        }


        internal void Dispose() {
            LoggedIn = false;

            ((AresTcpSocket)socket).Dispose();

            socket = null;

            guid = Guid.Empty;
            name = null;
            orgname = null;
            version = null;
            region = null;
            message = null;
            avatar = null;
            orgavatar = null;
            localip = null;
            nodeip = null;
            dnsentry = null;

            counters.Clear();
            counters = null;

            files.Clear();
            files = null;
            pfiles = null;

            ignored.Clear();
            ignored = null;

            extended_props.Clear();
            extended_props = null;
        }

        //we don't want a plugin to be able
        //to dispose us directly... bad
        void IDisposable.Dispose() {
            if (Connected) Disconnect();
        }

        public override int GetHashCode() {
            if (cookie == 0)
                cookie = (uint)base.GetHashCode();

            return (int)cookie;
        }


        #region " INotifyPropertyChanged "

        protected virtual void RaisePropertyChanged<T>(Expression<Func<T>> property) {
            var prop = ((MemberExpression)property.Body).Member as PropertyInfo;

            if (prop == null) return;
            OnPropertyChanged(prop.Name);
        }

        protected virtual void RaisePropertyChanged(string propName) {
            OnPropertyChanged(propName);
        }

        protected virtual void OnPropertyChanged(string propName) {
            var x = PropertyChanged;
            if (x != null) x(this, new PropertyChangedEventArgs(propName));
        }

        public event PropertyChangedEventHandler PropertyChanged;

        #endregion
    }
}
