﻿using System;
using System.Net;
using System.Linq.Expressions;
using System.Reflection;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Threading;
using System.ComponentModel;

using Zorbo.Resources;

using Zorbo.Sockets;

using Zorbo.Packets;
using Zorbo.Packets.Ares;
using Zorbo.Packets.Formatters;

using Zorbo.Interface;
using Zorbo.Plugins;
using Zorbo.Users;

namespace Zorbo
{
    public class AresServer : IServer
    {
        Timer timer;
        TimeSpan ticklength;

        IServerConfig config;

        History history;
        AresChannels channels;
        AresTcpSocket listener;

        IPAddress localip;
        IPAddress externalip;

        PluginHost plugins;

        SortedStack<UInt16> idpool;
        Comparison<IClient> sorter = (a, b) => (a.Id - b.Id);

        List<PendingConnection> pending;
        List<IFloodRule> flood_rules;

        ObservableCollection<IClient> users;
        ReadOnlyList<IClient> plugin_users;

        class PendingConnection
        {
            public ISocket Socket { get; set; }
            public DateTime Time { get; set; }

            public PendingConnection() { }

            public PendingConnection(ISocket socket, DateTime time) {
                this.Socket = socket;
                this.Time = time;
            }
        }

        public Boolean Running {
            get { return listener != null; }
        }

        public IPAddress LocalIp {
            get { return localip; }
            private set {
                if (localip == null || !localip.Equals(value)) {
                    localip = value;
                    RaisePropertyChanged(() => LocalIp);

                    Logging.WriteLines(new[] { 
                        "----------",
                        String.Format("Reported local ip: {0}", value),
                        "----------"
                    });
                }
            }
        }

        public IPAddress ExternalIp {
            get { return externalip; }
            internal set {
                if (externalip == null || !externalip.Equals(value)) {

                    externalip = value;
                    RaisePropertyChanged(() => ExternalIp);

                    Logging.WriteLines(new[] { 
                        "----------",
                        String.Format("Reported external ip: {0}", value),
                        "----------"
                    });
                }
            }
        }

        public IServerConfig Config {
            get { return config; }
        }

        public IChannelList Channels {
            get { return channels; }
        }

        public IHistory History {
            get { return history; }
        }

        public IPluginHost PluginHost {
            get { return plugins; }
        }


        public IReadOnlyList<IClient> Users {
            get { return plugin_users; }
        }

        public IList<IFloodRule> FloodRules {
            get { return flood_rules; }
        }

        internal PluginHost PluginManager {
            get { return plugins; }
        }


        public AresServer(IServerConfig config) {
            this.config = config;

            ticklength = TimeSpan.FromSeconds(1);

            history = new History(this);
            channels = new AresChannels(this);
            plugins = new PluginHost(this);

            idpool = new SortedStack<UInt16>();
            idpool.SetSort((a, b) => (a - b));

            pending = new List<PendingConnection>();

            users = new ObservableCollection<IClient>();
            plugin_users = new ReadOnlyList<IClient>(users);

            flood_rules = new List<IFloodRule>();

            history.Load();
        }

        public void Start() {
            channels.Start();

            for (ushort i = 0; i < config.MaxClients; i++)
                idpool.Push(i);

            config.PropertyChanged += Config_PropertyChanged;

            listener = new AresTcpSocket(new ClientFormatter());
            listener.Accepted += ClientAccepted;

            Logging.WriteLines(new[] { 
                "Server started",
                "----------",
                String.Format("Listening on port: {0}", config.Port),
                "----------"
            });

            IPHostEntry entry = Dns.GetHostEntry(Environment.MachineName);

            foreach (var ip in entry.AddressList) {
                if (ip.AddressFamily == listener.Socket.AddressFamily) {
                    LocalIp = ip;
                    break;
                }
            }

            listener.Bind(new IPEndPoint(IPAddress.Any, config.Port));
            listener.Listen(25);

            timer = new Timer(new TimerCallback(OnTimer), null, ticklength, ticklength);
            RaisePropertyChanged(() => Running);
        }

        public void Stop() {

            timer.Change(-1, -1);
            timer.Dispose();

            config.PropertyChanged -= Config_PropertyChanged;

            listener.Dispose();
            listener = null;

            idpool.Clear();
            channels.Stop();

            RaisePropertyChanged(() => Running);

            lock (pending) {
                foreach (var sock in pending)
                    sock.Socket.Disconnect();

                pending.Clear();
            }

            lock (((ICollection)users).SyncRoot) {
                foreach (var user in users) {
                    history.Add(user);
                    ((AresClient)user).Dispose();
                }
                users.Clear();
            }

            history.Save();
            users.Clear();
        }


        private void OnTimer(object state) {
            DateTime now = TimeBank.CurrentTime;

            lock (pending) {
                foreach (var sock in pending) {
                    if (now.Subtract(sock.Time).TotalSeconds >= 30)
                        sock.Socket.Disconnect();
                }
            }

            lock (((ICollection)users).SyncRoot) {
                
                foreach (var user in users) {
                    AresClient client = (AresClient)user;

                    if (client.IsCaptcha) {
                        if (now.Subtract(client.CaptchaTime).TotalHours >= 2) {

                            SendAnnounce(client, Strings.BannedCaptchaTimeout);
                            client.Ban();
                            continue;
                        }
                    }

                    if (now.Subtract(client.LastUpdate).TotalMinutes >= 5)
                        client.Disconnect();
                }
            }
        }


        public void SendUserlist(IClient user) {

            lock (((ICollection)users).SyncRoot) {
                Join join = new Join(user);

                if (!user.IsCaptcha) {

                    user.Socket.SendAsync(new Userlist() { 
                        Username = "Zorbo",
                        FileCount = 420,
                        Level = AdminLevel.Host,
                    });

                    foreach (var s in users) {
                        if (s != user &&
                            s.Connected &&
                            s.LoggedIn &&
                           !s.IsCaptcha &&
                            s.Vroom == user.Vroom) {

                            s.Socket.SendAsync(join);
                            user.Socket.SendAsync(new Userlist(s));
                        }
                    }
                }

                user.Socket.SendAsync(new Userlist(user));
                user.Socket.SendAsync(new UserlistEnd());
            }
        }

        public void SendAvatars(IClient user) {

            lock (((ICollection)users).SyncRoot) {

                if (!Config.Avatar.Equals(AresAvatar.Null))
                    user.Socket.SendAsync(new ServerAvatar("Zorbo", Config.Avatar));

                user.Socket.SendAsync(new ServerPersonal("Zorbo", Strings.VersionLogin));

                foreach (var s in users) {

                    if (s.Connected &&
                        s.LoggedIn &&
                       !s.IsCaptcha &&
                        s.Vroom == user.Vroom) {

                        s.Socket.SendAsync(new ServerAvatar(user));

                        if (!String.IsNullOrEmpty(user.Message))
                            s.Socket.SendAsync(new ServerPersonal(user.Name, user.Message));


                        user.Socket.SendAsync(new ServerAvatar(s));

                        if (!String.IsNullOrEmpty(s.Message))
                            user.Socket.SendAsync(new ServerPersonal(s.Name, s.Message));
                    }
                }
            }
        }


        public void SendPacket(IPacket packet) {
            lock (((ICollection)users).SyncRoot) {
                foreach (var user in users)
                    if (user.Connected && user.LoggedIn && !user.IsCaptcha)
                        user.Socket.SendAsync(packet);
            }
        }

        public void SendPacket(IClient user, IPacket packet) {
            if (user.Connected && user.LoggedIn) 
                user.Socket.SendAsync(packet);
        }

        public void SendPacket(Predicate<IClient> match, IPacket packet) {
            lock (((ICollection)users).SyncRoot) {
                foreach (var user in users)
                    if (user.Connected && 
                        user.LoggedIn &&
                       !user.IsCaptcha &&
                        match(user)) user.Socket.SendAsync(packet);
            }
        }


        public IClient FindUser(Predicate<IClient> match) {
            lock (((ICollection)users).SyncRoot) {
                foreach (var user in users)
                    if (user.Connected && 
                        user.LoggedIn && 
                        match(user)) return user;
            }

            return null;
        }


        protected virtual void ClientAccepted(object sender, AcceptEventArgs e) {

            var socket = (AresTcpSocket)e.Socket;

            if (!channels.FinishedTestingFirewall && 
                 channels.IsCheckingMyFirewall(socket.RemoteEndPoint)) {

                socket.Dispose();
            }
            else {
                socket.Exception += ClientException;
                socket.Disconnected += ClientDisconnected;
                socket.PacketReceived += ClientPacketReceived;

                lock(pending) pending.Add(new PendingConnection(socket, TimeBank.CurrentTime));

                socket.ReceiveAsync();
            }
        }

        protected virtual void ClientException(object sender, ExceptionEventArgs e) {

            AresClient user = null;

            lock (((ICollection)users).SyncRoot)
                user = (AresClient)users.Find(s => s.Socket == sender);

            if (user == null) return;

            Logging.WriteLines(new[] {
                "----------",
                String.Format("Exception occured in client: {0}{1}Message: {2}", user.Name, Environment.NewLine, e.Exception.Message),
                "----------"
            });

            user.Disconnect();
        }

        protected virtual void ClientPacketReceived(object sender, PacketEventArgs e) {
            if (e.Packet != null) {
                
                AresClient user = null;
                AresTcpSocket socket = (AresTcpSocket)sender;

                try {
                    lock (((ICollection)users).SyncRoot)
                        user = (AresClient)users.Find(s => s.Socket == socket);

                    if ((AresId)e.Packet.Id == AresId.MSG_CHAT_CLIENT_LOGIN) {

                        if (user == null) {
                            lock (pending) pending.RemoveAll((s) => s.Socket == socket);

                            int id = idpool.Pop();
                            user = new AresClient(this, socket, (ushort)id);

                            if (user.ExternalIp.IsLocal() || user.ExternalIp.Equals(ExternalIp))
                                user.LocalHost = true;

                            lock (((ICollection)users).SyncRoot) {
                                users.Add(user);
                                users.Sort(sorter);
                            }

                            user.HandleJoin(e);
                        }
                        else {
                            user.Socket.SendAsync(new Announce(Errors.LoginFlood));
                            user.Socket.Disconnect();
                        }
                    }
                    else if (user != null && user.LoggedIn && CheckCounters(user, e.Packet)) {

                        user.LastUpdate = TimeBank.CurrentTime;

                        bool @override = !plugins.OnBeforePacket(user, e.Packet);

                        if (!user.HandlePacket(e) && !@override)
                            user.HandleOverridePacket(e);

                        plugins.OnAfterPacket(user, e.Packet);
                    }
                }
                catch(Exception ex) {

                    Logging.WriteLines(new[] { 
                        "----------",
                        String.Format("Exception occured: {0}", ex.Message),
                        "----------"
                    });

                    if (user != null)
                        user.Disconnect();
                    else
                        socket.Disconnect();
                }
            }
        }

        protected virtual void ClientDisconnected(object sender, DisconnectEventArgs e) {
            AresClient user = null;
            AresTcpSocket socket = (AresTcpSocket)sender;

            socket.Exception -= ClientException;
            socket.Disconnected -= ClientDisconnected;
            socket.PacketReceived -= ClientPacketReceived;

            lock (pending) pending.RemoveAll((s) => s.Socket == socket);

            lock (((ICollection)users).SyncRoot) {

                int uindex = users.FindIndex(s => s.Socket == sender);
                if (uindex == -1) return;

                user = (AresClient)users[uindex];
                users.RemoveAt(uindex);

                if (user.LoggedIn) {
                    history.Add(user);

                    SendPacket((s) =>
                        !s.Guid.Equals(user.Guid) && //ghost?
                         s.Vroom == user.Vroom, 
                         new Parted(user.Name));
                }
            }

            plugins.OnPart(user, e.UserToken);
            idpool.Push(user.Id);

            user.Dispose();
        }

        protected virtual bool CheckCounters(AresClient user, IPacket packet) {

            var rules = flood_rules.FindAll((s) => s.Id == packet.Id);

            foreach(var rule in rules) {

                Counter counter = null;
                DateTime now = TimeBank.CurrentTime;

                if (!user.Counters.ContainsKey(packet.Id)) {

                    counter = new Counter(0, now);
                    user.Counters.Add(packet.Id, counter);
                }
                else {
                    counter = user.Counters[packet.Id];

                    if (now.Subtract(counter.Last).TotalMilliseconds > rule.ResetTimeout)
                        counter.Count = 0;
                }

                if (++counter.Count >= rule.Count) {

                    if (!plugins.OnFlood(user, packet))
                        return false;
                }
                else counter.Last = now;
            }
            
            return true;
        }


        private void Config_PropertyChanged(object sender, PropertyChangedEventArgs e) {

            switch (e.PropertyName) {
                case "Topic":
                    SendPacket(new Topic(Config.Topic));
                    break;
            }
        }


        #region " IServer Methods "

        public void SendText(string target, string sender, string text) {
            SendPacket((s) => s.Name == target, new ServerPublic(sender, text));
        }

        public void SendText(IClient target, string sender, string text) {
            SendPacket((s) => s == target, new ServerPublic(sender, text));
        }

        public void SendText(IClient target, IClient sender, string text) {
            SendPacket((s) => s == target, new ServerPublic(sender.Name, text));
        }

        public void SendText(string sender, string text) {
            SendPacket(new ServerPublic(sender, text));
        }

        public void SendText(IClient sender, string text) {
            SendPacket(new ServerPublic(sender.Name, text));
        }

        public void SendText(Predicate<IClient> match, string sender, string text) {
            SendPacket(match, new ServerPublic(sender, text));
        }

        public void SendText(Predicate<IClient> match, IClient sender, string text) {
            SendPacket(match, new ServerPublic(sender.Name, text));
        }


        public void SendEmote(string target, string sender, string text) {
            SendPacket((s) => s.Name == target, new ServerEmote(sender, text));
        }

        public void SendEmote(IClient target, string sender, string text) {
            SendPacket((s) => s == target, new ServerEmote(sender, text));
        }

        public void SendEmote(IClient target, IClient sender, string text) {
            SendPacket((s) => s == target, new ServerEmote(sender.Name, text));
        }

        public void SendEmote(string sender, string text) {
            SendPacket(new ServerEmote(sender, text));
        }

        public void SendEmote(IClient sender, string text) {
            SendPacket(new ServerEmote(sender.Name, text));
        }

        public void SendEmote(Predicate<IClient> match, string sender, string text) {
            SendPacket(match, new ServerEmote(sender, text));
        }

        public void SendEmote(Predicate<IClient> match, IClient sender, string text) {
            SendPacket(match, new ServerEmote(sender.Name, text));
        }


        public void SendPrivate(string target, string sender, string text) {
            SendPacket((s) => s.Name == target, new Private(sender, text));
        }

        public void SendPrivate(IClient target, string sender, string text) {
            SendPacket((s) => s == target, new Private(sender, text));
        }

        public void SendPrivate(IClient target, IClient sender, string text) {
            SendPacket((s) => s == target, new Private(sender.Name, text));
        }

        public void SendPrivate(Predicate<IClient> match, string sender, string text) {
            SendPacket(match, new Private(sender, text));
        }

        public void SendPrivate(Predicate<IClient> match, IClient sender, string text) {
            SendPacket(match, new Private(sender.Name, text));
        }


        public void SendAnnounce(string text) {
            SendPacket(new Announce(text));
        }

        public void SendAnnounce(string target, string text) {
            SendPacket((s) => s.Name == target, new Announce(text));
        }

        public void SendAnnounce(IClient target, string text) {
            SendPacket((s) => s == target, new Announce(text));
        }

        public void SendAnnounce(Predicate<IClient> match, string text) {
            SendPacket(match, new Announce(text));
        }


        public void SendWebsite(string address, string caption) {
            SendPacket(new Website(address, caption));
        }

        public void SendWebsite(string target, string address, string caption) {
            SendPacket((s) => s.Name == target, new Website(address, caption));
        }

        public void SendWebsite(IClient target, string address, string caption) {
            SendPacket((s) => s == target, new Website(address, caption));
        }

        public void SendWebsite(Predicate<IClient> match, string address, string caption) {
            SendPacket(match, new Website(address, caption));
        }

        #endregion

        #region " INotifyPropertyChanged "

        protected virtual void RaisePropertyChanged<T>(Expression<Func<T>> property) {
            var prop = ((MemberExpression)property.Body).Member as PropertyInfo;

            if (prop == null) return;
            OnPropertyChanged(prop.Name);
        }

        protected virtual void RaisePropertyChanged(string propName) {
            OnPropertyChanged(propName);
        }

        protected virtual void OnPropertyChanged(string propName) {
            var x = PropertyChanged;
            if (x != null) x(this, new PropertyChangedEventArgs(propName));
        }

        public event PropertyChangedEventHandler PropertyChanged;

        #endregion
    }
}
